#!/usr/local/bin/ruby
#
# Daedalus provides a simple mechanism for running programs in response to the
# output of "check programs".  In practice this is generally used to provide an
# automated restart service for programs by chacking the output of "ps ax" and
# running a start script when a particular program fails to appear.
#
# Author:: David Powers
# Copyright:: November, 2004
# License:: BSD


require 'logger'
require 'optparse'
require 'rexml/document'
require 'timeout'


# raised for configuration file errors
class ConfigurationError < Exception
end


class ConfigurationFileReader
	
	# reads in a configuration from the given filename and returns it as a Configuration object
	def read(filename)
		xmldoc = nil
		File.open(filename, File::RDONLY) { |fp|
			xmldoc = REXML::Document.new(fp)
		}
		
		@configuration = Configuration.new()
		
		process_templates(xmldoc.root.elements['templates'])
		process_monitors(xmldoc.root.elements['monitors'])
		
		return(@configuration)
	end
	
	
	protected
	
	# returns either the setting from the xml or default.  conversionmethod is applied to the
	# setting to ensure the final return value is of a known type.  If a block is provided setting will
	# be passed to the block (after conversion) for constraint checking or modification.
	def get_setting(xml, default, conversionmethod = :to_s)
		setting = nil
		
		if(xml != nil)
			setting = xml.text()
		elsif(default != nil)
			setting = default
		end
		
		# return if setting is still nil - that means nothing was in the conf file
		# and no meaningful default was given.  The caller likely wants it to return
		# nil
		if(setting == nil)
			return(setting)
		end
		
		# convert based on the conversion method
		setting = setting.method(conversionmethod).call()
		
		if(block_given?())
			yield(setting)
		end
		
		return(setting)
	end
	
	
	# Handles the <templates> section of the configuration file
	def process_templates(xml)
		if(xml == nil)
			return
		end
		
		xml.each_element() { |templatexml|
			template = Hash.new()
			
			template['name'] = get_setting(templatexml.elements['name'], nil)
			if(template['name'] == nil)
				raise(ConfigurationError, "All templates must have a name")
			end
			existancecheck = false
			@configuration.templates.each() { |templatecheck|
				if(templatecheck['name'] == template['name'])
					raise(ConfigurationError, "All templates must have a unique name - \"#{template['name']}\" has been used at least twice")
				end
			}
			
			template['regex'] = get_setting(templatexml.elements['regex'], nil)
			if(template['regex'] != nil)
				if(template['regex'] =~ /^\/.*\/$/)
					template['regex'] = Regexp.new(template['regex'][/^\/(.*)\/$/, 1])
				else
					template['regex'] = Regexp.new(Regexp.escape(template['regex']))
				end
			end
			
			template['logtype'] = get_setting(templatexml.elements['logtype'], nil) { |logtype|
				if(logtype != 'normal' and logtype != 'statechangeonly')
					raise(ConfigurationError, "<logtype> must be set to normal or statechangeonly - is currently set to \"#{logtype}\"")
				end
			}
			
			template['checkcommand'] = get_setting(templatexml.elements['checkcommand'], nil)
			template['failcommand'] = get_setting(templatexml.elements['failcommand'], nil)
			template['successcommand'] = get_setting(templatexml.elements['successcommand'], nil)
			template['failalertcommand'] = get_setting(templatexml.elements['failalertcommand'], nil)
			template['successalertcommand'] = get_setting(templatexml.elements['successalertcommand'], nil)
			template['checkinterval'] = get_setting(templatexml.elements['checkinterval'], nil, :to_i)
			template['commandtimeout'] = get_setting(templatexml.elements['commandtimeout'], nil, :to_i)
			template['template'] = get_setting(templatexml.elements['template'], nil)
			
			@configuration.templates << template 
		}
	end
	
	
	# Handles the <monitors> section of the configuration file
	def process_monitors(xml)
		if(xml == nil)
			return
		end
		
		xml.each_element() { |monitorxml|
			monitor = Hash.new()
			monitor['name'] = get_setting(monitorxml.elements['name'], nil)
			if(monitor['name'] == nil)
				raise(ConfigurationError, "All monitors must have a name")
			end
			
			monitor['regex'] = get_setting(monitorxml.elements['regex'], nil)
			if(monitor['regex'] != nil)
				if(monitor['regex'] =~ /^\/.*\/$/)
					monitor['regex'] = Regexp.new(monitor['regex'][/^\/(.*)\/$/, 1])
				else
					monitor['regex'] = Regexp.new(Regexp.escape(monitor['regex']))
				end
			end
			
			monitor['logtype'] = get_setting(monitorxml.elements['logtype'], nil) { |logtype|
				if(logtype != 'normal' and logtype != 'statechangeonly')
					raise(ConfigurationError, "<logtype> must be set to normal or statechangeonly - is currently set to \"#{logtype}\"")
				end
			}
			
			monitor['checkcommand'] = get_setting(monitorxml.elements['checkcommand'], nil)
			monitor['failcommand'] = get_setting(monitorxml.elements['failcommand'], nil)
			monitor['successcommand'] = get_setting(monitorxml.elements['successcommand'], nil)
			monitor['failalertcommand'] = get_setting(monitorxml.elements['failalertcommand'], nil)
			monitor['successalertcommand'] = get_setting(monitorxml.elements['successalertcommand'], nil)
			monitor['checkinterval'] = get_setting(monitorxml.elements['checkinterval'], nil, :to_i)
			monitor['commandtimeout'] = get_setting(monitorxml.elements['commandtimeout'], nil, :to_i)
			
			monitor['template'] = get_setting(monitorxml.elements['template'], nil) 
			
			@configuration.monitors << monitor
		}
	end

end



# Holds an internal representation of the configuration
class Configuration
	attr_accessor :templates, :monitors
	
	def initialize()
		@templates = Array.new() # holds the templates in load order for proper cascading
		@monitors = Array.new()
	end
	
end



# Represents the monitor of a single service
class Monitor
	
	def initialize(configuration)
		@configuration = configuration
		@state = "unknown"
	end
	
	
	# starts the monitor monitoring
	def start()
		@thread = Thread.new() {
			@state = 'unknown'
			while(true)
				$logger.debug("running \"#{@configuration['checkcommand']}\" for monitor \"#{@configuration['name']}\"")
				output = `#{@configuration['checkcommand']}`
				
				match = @configuration['regex'].match(output)
				changed = set_state(match)
				if(match == nil and @configuration['failcommand'] != nil)
					if(@configuration['failalertcommand'] != nil)
						run_alert(@configuration['failalertcommand'], (changed == true or @configuration['logtype'] != 'statechangeonly'))
					end
					
					if(changed == true or @configuration['logtype'] != 'statechangeonly')
						$logger.warn("Regular expression /#{@configuration['regex'].source}/ did not match the output of \"#{@configuration['checkcommand']}\" for monitor \"#{@configuration['name']}\"")
					end
					begin
						if(changed == true or @configuration['logtype'] != 'statechangeonly')
							$logger.warn("Running \"#{@configuration['failcommand']}\" (failcommand) for monitor \"#{@configuration['name']}\"")
						end
						timeout(@configuration['commandtimeout']) {
							success = system(@configuration['failcommand'])
							if(!success)
								$logger.error("Command \"#{@configuration['failcommand']}\" did not return successfully")
							end
						}
					rescue TimeoutError
						$logger.error("Command \"#{@configuration['failcommand']}\" failed to execute within the timeout period")
					end
				elsif(match != nil and @configuration['successcommand'])
					if(@configuration['successalertcommand'] != nil)
						run_alert(@configuration['successalertcommand'], (changed == true or @configuration['logtype'] != 'statechangeonly'))
					end
					
					if(changed == true or (@configuration['logtype'] != 'statechangeonly' and @configuration['successcommand'] != nil))
						$logger.warn("Regular expression /#{@configuration['regex'].source}/ matched the output of \"#{@configuration['checkcommand']}\" for monitor \"#{@configuration['name']}\"")
					end
					begin
						if(changed == true or @configuration['logtype'] != 'statechangeonly')
							$logger.warn("Running \"#{@configuration['successcommand']}\" (successcommand) for monitor \"#{@configuration['name']}\"")
						end
						timeout(@configuration['commandtimeout']) {
							success = system(@configuration['successcommand'])
							if(!success)
								$logger.error("Command \"#{@configuration['successcommand']}\" did not return successfully")
							end
						}
					rescue TimeoutError
						$logger.error("Command \"#{@configuration['successcommand']}\" failed to execute within the timeout period")
					end
				else
					$logger.debug("No success or fail commands need to be run for monitor \"#{@configuration['name']}\"")
				end
				
				sleep(@configuration['checkinterval'])
			end
		}
	end
	
	
	# stops the monitor safely (we hope)
	def stop()
		@thread.stop()
	end
	
	
	protected
	
	
	# sets the state of the monitor (state is determined to be found if newstate evalutates to false or nil
	# and not found otherwise).  Returns true if the state actually changed.
	def set_state(newstate)
		if(!newstate)
			if(@state != "not found")
				$logger.info("State changed from \"#{@state}\" on monitor #{@configuration['name']} - the new state is \"not found\"")
				@state = "not found"
				return(true)
			end
		else
			if(@state != "found")
				$logger.info("State changed from \"#{@state}\" on monitor #{@configuration['name']} - the new state is \"found\"")
				@state = "found"
				return(true)
			end
		end
		
		return(false)
	end
	
	
	# Runs the given alert command in a seperate thread to avoid blocking
	def run_alert(alertcommand, log=true)
		Thread.new() {
			begin
				timeout(@configuration['commandtimeout']) {
					if(log)
						$logger.warn("Running \"#{alertcommand}\" (alert command) for monitor \"#{@configuration['name']}\"")
					end
					success = system(alertcommand)
					if(!success)
						$logger.error("Command \"#{alertcommand}\" did not return successfully")
					end
				}
			rescue TimeoutError
				$logger.error("Command \"#{alertcommand}\" failed to execute within the timeout period")
			end
		}
	end
end



class Template
	
	def initialize(configuration)
		@configuration = configuration
	end
	
	
	def apply(monitorconfiguration)
		$logger.debug("Applying template \"#{@configuration['name']}\" to monitorconfiguration \"#{monitorconfiguration['name']}\"")
		
		@configuration.each() { |key, value|
			if(monitorconfiguration[key] == nil)
				$logger.debug("Adding #{key} = \"#{value}\" to monitorconfiguration \"#{monitorconfiguration['name']}\" from template \"#{@configuration['name']}\"")
				monitorconfiguration[key] = value
			end
		}
		
		return(monitorconfiguration)
	end
	
end



module Daedalus

	# Sets configuration defaults and overrides from the command line
	def Daedalus::get_options()
		options = Hash.new()
		options['config'] = 'daedalus.xml'
		options['log'] = '/var/log/daedalus.log'
		options['verbosity'] = Logger::INFO
		options['pidfile'] = '/var/run/daedalus.pid'
		options['daemon'] = true
		
		opts = OptionParser.new() { |opts|
			opts.on_tail("-v", "--verbosity VERBOSITY", Integer, "(0-4) 0 being fatal errors only, default 3") { |verbosity|
				if(verbosity == 0)
					options['verbosity'] = Logger::FATAL
				elsif(verbosity == 1)
					options['verbosity'] = Logger::ERROR
				elsif(verbosity == 2)
					options['verbosity'] = Logger::WARN
				elsif(verbosity == 3)
					options['verbosity'] = Logger::INFO
				elsif(verbosity == 4)
					options['verbosity'] = Logger::DEBUG
				else
					$stderr.print("verbosity must be set between 0 and 4\n")
					exit(1)
				end
			}
			
			opts.on_tail("-h", "--help", "Print this message") {
				print(opts)
				exit()
			}
			
			opts.on("-t", "--templates TEMPLATEFILE", "Location of the template file") { |filename|
				options['templates'] = filename
			}

			opts.on("-c", "--config CONFIGFILE", "Location of the configuration file") { |filename|
				options['config'] = filename
			}
			
			opts.on("-l", "--log LOGFILE", "Location of the log file") { |filename|
				options['log'] = filename
			}
			
			opts.on("-p", "--pidfile PIDFILE", "Location of the pid file") { |filename|
				options['pidfile'] = filename
			}
			
			opts.on("-d", "--nodaemon", "Do not fork into the background") {
				options['daemon'] = false
			}
		}
		
		opts.parse(ARGV)
		
		return(options)
	end
	
	
	# forks the program into the background and detaches it from the terminal
	def Daedalus.daemonize()
		# fork into the background
		# Parent exits, child continues.
		if(fork())
			exit()
		end
		
		# Become session leader.
		Process.setsid()
		
		# Zap session leader.
		if(fork())
			exit()
		end
		
		# Release old working directory.
		Dir.chdir "/"
		
		# Ensure sensible umask. Adjust as needed.
		File.umask(0000)
		
		# reopen the file descriptors to nothing
		STDIN.reopen('/dev/null')
		STDOUT.reopen('/dev/null', 'a')
		STDERR.reopen(STDOUT)
	end
	
	
	# writes the PID to a file
	def Daedalus.writepid(pidfile)
		File.open(pidfile, File::WRONLY|File::CREAT|File::TRUNC, 0644) { |fp|
			fp.write("#{Process.pid}\n")
		}
	end
	
	
	# Main processing loop
	def Daedalus::main()
		options = get_options()
		
		# create a logger
		$logger = Logger.new(options['log'], 10, 1024000)
		$logger.level = options['verbosity']
		$logger.info("daedalus started")
		
		# setup some monitor configuration defaults
		defaults = Hash.new()
		defaults['commandtimeout'] = 60
		defaults['logtype'] = 'normal'
		
		# read in templates if any are definied in a separate file
		templateconfiguration = nil
		if(options['templates'] != nil)
			templateconfiguration = ConfigurationFileReader.new().read(options['templates'])
		end
		
		# read in the configuration
		configuration = ConfigurationFileReader.new().read(options['config'])
		
		# merge the templates (if they exist) with the configuration templates
		if(templateconfiguration != nil)
			configuration.templates = templateconfiguration.templates + configuration.templates
		end
		
		# create the template objects - apply templates to template configurations
		# as we go so that we end up with a cascading template effect
		templates = Hash.new()
		configuration.templates.each() { |templateconfiguration|
			if(templateconfiguration['template'] != nil)
				if(templates[templateconfiguration['template']] == nil)
					raise(ConfigurationError, "Template \"#{templateconfiguration['name']}\" references a template \"#{tempndelateconfiguration['template']}\" that does not defined")
				end
				templateconfiguration = templates[templateconfiguration['template']].apply(templateconfiguration)
			end
			
			$logger.debug("Creating template \"#{templateconfiguration['name']}\"")
			templates[templateconfiguration['name']] = Template.new(templateconfiguration)
		}
		
		# create the monitor objects
		monitors = Array.new()
		configuration.monitors.each() { |monitorconfiguration|
			if(monitorconfiguration['template'] != nil)
				if(templates[monitorconfiguration['template']] == nil)
					raise(ConfigurationError, "Monitor \"#{monitorconfiguration['name']}\" references a template \"#{monitorconfiguration['template']}\" that does not exist")
				end
				
				monitorconfiguration = templates[monitorconfiguration['template']].apply(monitorconfiguration)
			end
			
			defaults.each() { |key, value|
				if(monitorconfiguration[key] == nil)
					$logger.debug("Adding #{key} = \"#{value}\" to monitorconfiguration \"#{monitorconfiguration['name']}\" from defaults")
					monitorconfiguration[key] = value
				end
			}

			monitors << Monitor.new(monitorconfiguration)
		}
		
		if(options['daemon'])
			daemonize()
		end
		writepid(options['pidfile'])
		
		# start all the monitors
		monitors.each() { |monitor|
			monitor.start()
		}
		
		while(true)
			sleep(10)
		end
	end

end


Daedalus.main()